import struct

from seabreeze.pyseabreeze.protocol import OOIProtocol, OBPProtocol
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
class SeaBreezeThermoElectricFeature(SeaBreezeFeature):
    identifier = 'thermo_electric'

    def read_temperature_degrees_celsius(self):
        raise NotImplementedError("implement in derived class")

    def set_temperature_setpoint_degrees_celsius(self, temperature):
        raise NotImplementedError("implement in derived class")

    def enable_tec(self, state):
        raise NotImplementedError("implement in derived class")


# OOI implementation
# ==================
#
class ThermoElectricFeatureOOI(SeaBreezeThermoElectricFeature):
    _required_protocol_cls = OOIProtocol

    def enable_tec(self, state):
        self.protocol.send(0x71, (bool(state), 0x00))

    def set_temperature_setpoint_degrees_celsius(self, temperature):
        self.protocol.send(0x73, temperature * 10)

    def read_temperature_degrees_celsius(self):
        self.protocol.send(0x72)
        ret = self.protocol.receive(timeout_ms=1000, mode='low_speed')
        return struct.unpack('<h', ret)[0] / 10.0

    # NOTE: if implemented this would also need to be added to the interface
    # def set_tec_fan_enable(self, tec_fan_enable):
    #    # this is not implemented in the seabreeze library?
    #    self.usb_send(struct.pack('<BBB', 0x70, bool(tec_fan_enable), 0x00))
    #    raise NotImplementedError


# OBP implementation
# ==================
#
class ThermoElectricFeatureOBP(SeaBreezeThermoElectricFeature):
    _required_protocol_cls = OBPProtocol

    def enable_tec(self, state):
        self.protocol.send(0x00420010, bool(state))

    def set_temperature_setpoint_degrees_celsius(self, temperature):
        self.protocol.send(0x00420011, float(temperature))

    def read_temperature_degrees_celsius(self):
        ret = self.protocol.query(0x00420004)
        return struct.unpack("<f", ret)[0]

    # def set_tec_fan_enable(self, tec_fan_enable):
    #     raise NotImplementedError



from functools import partial

from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.protocol import ProtocolInterface

try:
    from functools import partialmethod
except ImportError:
    # https://gist.github.com/carymrobbins/8940382
    # noinspection PyPep8Naming
    class partialmethod(partial):
        def __get__(self, instance, owner):
            if instance is None:
                return self
            args, kwargs = self.args or (), self.keywords or {}
            return partial(self.func, instance, *args, **kwargs)


class SeaBreezeFeature(object):
    identifier = "base_feature"

    _required_kwargs = ()
    _required_features = ()
    _required_protocol_cls = ProtocolInterface

    def __init__(self, protocol, feature_id, **kwargs):
        """SeaBreezeFeature base class

        Parameters
        ----------
        protocol : Type[seabreeze.pyseabreeze.protocol.ProtocolInterface}
        feature_id : int
        """
        if self.identifier == "base_feature":
            raise SeaBreezeError("Don't instantiate SeaBreezeFeature directly. Use derived feature classes.")
        assert set(self._required_kwargs) == set(kwargs), "{} vs {}".format(
            str(set(self._required_kwargs)), str(set(kwargs))
        )
        # check protocol support
        if not isinstance(protocol, self._required_protocol_cls):
            raise SeaBreezeError("FeatureError: Protocol not supported by feature")
        self.protocol = protocol
        self.feature_id = feature_id

    def __repr__(self):
        return "<{} id={}>".format(self.__class__.__name__, self.feature_id)

    @classmethod
    def get_feature_class_registry(cls):
        # noinspection PyUnresolvedReferences
        return {feature_class.identifier: feature_class for feature_class in SeaBreezeFeature.__subclasses__()}

    @classmethod
    def supports_protocol(cls, protocol):
        return isinstance(protocol, cls._required_protocol_cls)

    @classmethod
    def specialize(cls, model_name, **kwargs):
        assert set(kwargs) == set(cls._required_kwargs)
        specialized_class = type("{}{}".format(cls.__name__, model_name), (cls, ), {
            '__init__': partialmethod(cls.__init__, **kwargs)
        })
        return specialized_class



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeMulticastFeature(SeaBreezeFeature):
    identifier = 'multicast'

    def get_multicast_enable_state(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_multicast_enable_state(self, interface_index, enable_state):
        raise NotImplementedError("implement in derived class")



import pytest

try:
    import seabreeze.pyseabreeze as psb
except ImportError:
    psb = None

try:
    import seabreeze.cseabreeze as csb
except ImportError:
    csb = None


# skip all tests in this module if no backends can be imported
if csb is not None or psb is not None:
    import seabreeze
    seabreeze.use('cseabreeze', force=False)
    from seabreeze.spectrometers import list_devices, Spectrometer
else:
    pytestmark = pytest.mark.skip("No backend available!")


def _retr():
    """retrieves a list of all connected spectrometers for all backends"""
    params, ids = [], []
    csb_serials, psb_serials = set(), set()
    for serials, backend in [(csb_serials, csb), (psb_serials, psb)]:
        if backend is None:
            continue
        api = backend.SeaBreezeAPI()
        try:
            serials.update((d.serial_number, d.model) for d in api.list_devices())
        except:
            pass
        finally:
            api.shutdown()
    for ser_mod in csb_serials.union(psb_serials):
        params.extend([
            (csb, ser_mod[0]) if ser_mod in csb_serials else pytest.param((csb, None), marks=pytest.mark.skip),
            (psb, ser_mod[0]) if ser_mod in psb_serials else pytest.param((psb, None), marks=pytest.mark.skip),
        ])
        ids.extend([
            ("cseabreeze:{}:{}".format(ser_mod[1], ser_mod[0])),
            ("pyseabreeze:{}:{}".format(ser_mod[1], ser_mod[0])),
        ])
    if not params:
        params = [
            pytest.param((csb, None), marks=pytest.mark.skip),
            pytest.param((psb, None), marks=pytest.mark.skip),
        ]
        ids = [("cseabreeze:no-spectrometer"), ("pyseabreeze:no-spectrometer")]
    return dict(params=params, ids=ids)


@pytest.fixture(scope='function',
                params=[csb or pytest.param(csb, marks=pytest.mark.skip),
                        psb or pytest.param(psb, marks=pytest.mark.skip)],
                ids=["cseabreeze", "pyseabreeze"])
def backendlify(request):
    backend = request.param
    # list_devices
    _api = getattr(list_devices, '_api', None)
    if not isinstance(_api, backend.SeaBreezeAPI):
        _api = backend.SeaBreezeAPI()
        setattr(list_devices, '_api', _api)
    # Spectrometer
    Spectrometer._backend = backend
    try:
        yield
    finally:
        _api.shutdown()


@pytest.fixture(scope='function', **_retr())  # request all spectrometers on module load
def backendlified_serial(request):
    backend, serial = request.param
    # list_devices
    _api = getattr(list_devices, '_api', None)
    if not isinstance(_api, backend.SeaBreezeAPI):
        _api = backend.SeaBreezeAPI()
        setattr(list_devices, '_api', _api)
    # Spectrometer
    Spectrometer._backend = backend
    try:
        yield serial
    finally:
        _api.shutdown()


@pytest.mark.usefixtures('backendlify')
def test_cant_find_serial():
    exc = Spectrometer._backend.SeaBreezeError
    with pytest.raises(exc):
        Spectrometer.from_serial_number("i-do-not-exist")


def test_read_model(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    model = spec.model
    assert len(model) > 0


def test_read_serial_number(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    serial = spec.serial_number
    assert len(serial) > 0


def test_read_intensities(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    arr = spec.intensities()
    assert arr.size == spec.pixels


def test_read_wavelengths(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    arr = spec.wavelengths()
    assert arr.size == spec.pixels


def test_read_spectrum(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    arr = spec.spectrum()
    assert arr.shape == (2, spec.pixels)


def test_max_intensity(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    value = spec.max_intensity
    assert isinstance(value, float)
    assert 0 < value < 1e8


def test_integration_time_limits(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    spec = Spectrometer.from_serial_number(backendlified_serial)
    low, high = spec.integration_time_micros_limits
    assert isinstance(low, int)
    assert isinstance(high, int)
    assert 0 < low < high < 2**64


def test_integration_time(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    exc = Spectrometer._backend.SeaBreezeError
    spec = Spectrometer.from_serial_number(backendlified_serial)

    with pytest.raises(exc):
        # fail because too low
        spec.integration_time_micros(0)

    with pytest.raises(exc):
        # fail because too large
        spec.integration_time_micros(2**62)

    with pytest.raises(exc):
        # fail because too large long overflow
        spec.integration_time_micros(2**64)

    spec.integration_time_micros(10000)


def test_trigger_mode(backendlified_serial):
    devices = list(list_devices())
    if len(devices) == 0:
        pytest.skip("no supported device connected")

    exc = Spectrometer._backend.SeaBreezeError
    spec = Spectrometer.from_serial_number(backendlified_serial)

    spec.trigger_mode(0x00)  # <- normal mode

    with pytest.raises(exc):
        spec.trigger_mode(0xF0)  # <- should be unsupported for all specs
    # test again to see if the bus is locked
    spec.trigger_mode(0x00)  # <- normal mode




from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeWifiConfigurationFeature(SeaBreezeFeature):
    identifier = 'wifi_configuration'

    def get_wifi_mode(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_wifi_mode(self, interface_index, wifi_mode):
        raise NotImplementedError("implement in derived class")

    def get_wifi_security_type(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_wifi_security_type(self, interface_index, security_type):
        raise NotImplementedError("implement in derived class")

    def get_wifi_ssid(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_wifi_ssid(self, interface_index, ssid):
        raise NotImplementedError("implement in derived class")

    def set_wifi_pass_phrase(self, interface_index, pass_phrase):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeStrobeLampFeature(SeaBreezeFeature):
    identifier = 'strobe_lamp'

    def enable_lamp(self, state):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features.acquisitiondelay import SeaBreezeAcquisitionDelayFeature
from seabreeze.pyseabreeze.features.continuousstrobe import SeaBreezeContinuousStrobeFeature
from seabreeze.pyseabreeze.features.databuffer import SeaBreezeDataBufferFeature
from seabreeze.pyseabreeze.features.dhcpserver import SeaBreezeDHCPServerFeature
from seabreeze.pyseabreeze.features.eeprom import SeaBreezeEEPROMFeature
from seabreeze.pyseabreeze.features.ethernetconfiguration import SeaBreezeEthernetConfigurationFeature
from seabreeze.pyseabreeze.features.fastbuffer import SeaBreezeFastBufferFeature
from seabreeze.pyseabreeze.features.gpio import SeaBreezeGPIOFeature
from seabreeze.pyseabreeze.features.i2cmaster import SeaBreezeI2CMasterFeature
from seabreeze.pyseabreeze.features.introspection import SeaBreezeIntrospectionFeature
from seabreeze.pyseabreeze.features.ipv4 import SeaBreezeIPv4Feature
from seabreeze.pyseabreeze.features.irradcal import SeaBreezeIrradCalFeature
from seabreeze.pyseabreeze.features.lightsource import SeaBreezeLightSourceFeature
from seabreeze.pyseabreeze.features.multicast import SeaBreezeMulticastFeature
from seabreeze.pyseabreeze.features.networkconfiguration import SeaBreezeNetworkConfigurationFeature
from seabreeze.pyseabreeze.features.nonlinearity import SeaBreezeNonlinearityCoefficientsFeature
from seabreeze.pyseabreeze.features.opticalbench import SeaBreezeOpticalBenchFeature
from seabreeze.pyseabreeze.features.pixelbinning import SeaBreezePixelBinningFeature
from seabreeze.pyseabreeze.features.rawusb import SeaBreezeRawUSBBusAccessFeature
from seabreeze.pyseabreeze.features.revision import SeaBreezeRevisionFeature
from seabreeze.pyseabreeze.features.shutter import SeaBreezeShutterFeature
from seabreeze.pyseabreeze.features.spectrometer import SeaBreezeSpectrometerFeature
from seabreeze.pyseabreeze.features.spectrumprocessing import SeaBreezeSpectrumProcessingFeature
from seabreeze.pyseabreeze.features.straylightcoefficients import SeaBreezeStrayLightCoefficientsFeature
from seabreeze.pyseabreeze.features.strobelamp import SeaBreezeStrobeLampFeature
from seabreeze.pyseabreeze.features.temperature import SeaBreezeTemperatureFeature
from seabreeze.pyseabreeze.features.thermoelectric import SeaBreezeThermoElectricFeature
from seabreeze.pyseabreeze.features.wificonfiguration import SeaBreezeWifiConfigurationFeature
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeNetworkConfigurationFeature(SeaBreezeFeature):
    identifier = 'network_configuration'

    def get_number_of_network_interfaces(self):
        raise NotImplementedError("implement in derived class")

    def get_network_interface_connection_type(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def get_network_interface_enable_state(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_network_interface_enable_state(self, interface_index, enable_state):
        raise NotImplementedError("implement in derived class")

    def run_network_interface_self_test(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def save_network_interface_connection_settings(self, interface_index):
        raise NotImplementedError("implement in derived class")



#  tests for seabreeze-1.0.0-py38h6538335_0 (this is a generated file);
print('===== testing package: seabreeze-1.0.0-py38h6538335_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import os
import pytest

# workaround to undo conda-forge settng CI env var
os.environ['CI'] = 'true'
test_prefix = os.environ['PREFIX']

pytest.main(['-v', os.path.join(test_prefix, 'tests')])#  --- run_test.py (end) ---

print('===== seabreeze-1.0.0-py38h6538335_0 OK =====');
print("import: 'seabreeze'")
import seabreeze

print("import: 'seabreeze.cseabreeze'")
import seabreeze.cseabreeze

print("import: 'seabreeze.pyseabreeze'")
import seabreeze.pyseabreeze




"""This is a wrapper for the c-implementation of the seabreeze-library

"""
# noinspection PyUnresolvedReferences,PyProtectedMember
from seabreeze.cseabreeze._wrapper import (SeaBreezeAPI,
                                           SeaBreezeError,
                                           SeaBreezeNumFeaturesError,
                                           SeaBreezeDevice,
                                           SeaBreezeFeature,
                                           SeaBreezeRawUSBBusAccessFeature,
                                           SeaBreezeSpectrometerFeature,
                                           SeaBreezePixelBinningFeature,
                                           SeaBreezeThermoElectricFeature,
                                           SeaBreezeIrradCalFeature,
                                           SeaBreezeEthernetConfigurationFeature,
                                           SeaBreezeMulticastFeature,
                                           SeaBreezeIPv4Feature,
                                           SeaBreezeDHCPServerFeature,
                                           SeaBreezeNetworkConfigurationFeature,
                                           SeaBreezeWifiConfigurationFeature,
                                           SeaBreezeGPIOFeature,
                                           SeaBreezeEEPROMFeature,
                                           SeaBreezeLightSourceFeature,
                                           SeaBreezeStrobeLampFeature,
                                           SeaBreezeContinuousStrobeFeature,
                                           SeaBreezeShutterFeature,
                                           SeaBreezeNonlinearityCoefficientsFeature,
                                           SeaBreezeTemperatureFeature,
                                           SeaBreezeIntrospectionFeature,
                                           SeaBreezeSpectrumProcessingFeature,
                                           SeaBreezeRevisionFeature,
                                           SeaBreezeOpticalBenchFeature,
                                           SeaBreezeStrayLightCoefficientsFeature,
                                           SeaBreezeDataBufferFeature,
                                           SeaBreezeFastBufferFeature,
                                           SeaBreezeAcquisitionDelayFeature,
                                           SeaBreezeI2CMasterFeature)
_backend_ = 'cseabreeze'



import struct

import mock
import pytest

try:
    import seabreeze.pyseabreeze
except ImportError:
    pytestmark = pytest.mark.skip("Can't import pyseabreeze")
    OBP_VALID_MSG = None
else:
    from seabreeze.pyseabreeze.protocol import OBPProtocol
    _o = OBPProtocol.OBP
    OBP_VALID_MSG = struct.pack(
        _o.HEADER_FMT + "0s" + _o.FOOTER_FMT,
        _o.HEADER_START_BYTES,
        _o.HEADER_PROTOCOL_VERSION,
        0,
        _o.NO_ERROR,
        0x00000100,
        0,
        _o.RESERVED,
        _o.CHECKSUM_TYPE_NONE,
        1,
        b"c",
        20,
        b"",
        _o.NO_CHECKSUM,
        _o.FOOTER
    )


@pytest.fixture
def mock_transport():
    mock_transport = mock.Mock()
    mock_transport.read.return_value = OBP_VALID_MSG
    yield mock_transport


def test_obp_protocol_messages(mock_transport):
    obp = OBPProtocol(mock_transport)
    obp.send(0x00000100)




from seabreeze.pyseabreeze.protocol import ProtocolInterface
from seabreeze.pyseabreeze.transport import USBTransport
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


class SeaBreezeRawUSBBusAccessFeature(SeaBreezeFeature):
    identifier = 'raw_usb_bus_access'
    _required_protocol_cls = ProtocolInterface

    @classmethod
    def supports_protocol(cls, protocol):
        return isinstance(protocol.transport, USBTransport)

    def raw_usb_write(self, data, endpoint):
        if endpoint not in {'primary_out', 'secondary_out'}:
            raise ValueError("endpoint has to be one of 'primary_out', 'secondary_out'")
        if endpoint == 'secondary_out':
            pass
        self.protocol.transport.write(data)

    def raw_usb_read(self, endpoint, buffer_length=None):
        if endpoint == 'primary_in':
            return self.protocol.transport.read(size=buffer_length, mode='low_speed')
        elif endpoint == 'secondary_in':
            return self.protocol.transport.read(size=buffer_length, mode='high_speed')
        elif endpoint == 'secondary_in2':
            return self.protocol.transport.read(size=buffer_length, mode='high_speed_alt')
        else:
            raise ValueError("endpoint has to be one of 'primary_in', 'secondary_in', 'secondary_in2'")



import pytest


@pytest.fixture(scope='module')
def cseabreeze():
    yield pytest.importorskip('seabreeze.cseabreeze')


@pytest.fixture(scope='module')
def pyseabreeze():
    yield pytest.importorskip('seabreeze.pyseabreeze')


def _get_class_public_interface_dict(backend):
    base_class = backend.SeaBreezeFeature
    feature_classes = base_class.__subclasses__()
    interface_dict = {}
    for fcls in feature_classes:
        name = fcls.__name__
        attrs = set(attr for attr in dir(fcls) if not attr.startswith('_'))
        interface_dict[name] = attrs
    return interface_dict


def test_backend_features_interface(cseabreeze, pyseabreeze):
    """test if the public backend feature class interfaces are identical"""
    c_feature_interface = _get_class_public_interface_dict(cseabreeze)
    py_feature_interface = _get_class_public_interface_dict(pyseabreeze)

    assert set(c_feature_interface) == set(py_feature_interface)

    CSEABREEZE_CUSTOM = {'feature_id'}
    PYSEABREEZE_CUSTOM = {'supports_protocol', 'specialize'}

    for feature in c_feature_interface:
        c_attrs = c_feature_interface[feature] - CSEABREEZE_CUSTOM
        py_attrs = py_feature_interface[feature] - PYSEABREEZE_CUSTOM
        assert c_attrs == py_attrs, 'feature {} differs in attrs'.format(feature)


def test_cseabreeze_seabreezeapi(cseabreeze):
    """check if list_devices doesn't crash"""
    api = cseabreeze.SeaBreezeAPI()
    api.list_devices()


def test_pyseabreeze_seabreezeapi(pyseabreeze):
    """check if list_devices doesn't crash"""
    api = pyseabreeze.SeaBreezeAPI()
    api.list_devices()



import os
from distutils.util import strtobool

import mock
import pytest

# Fix for running pyseabreeze tests on CIs
RUNNING_ON_CI = strtobool(os.environ.get('CI', '0')) == 1


@pytest.fixture(scope="session", autouse=RUNNING_ON_CI)
def mock_pyusb_backend_libusb0():
    """mock usb.backend.libusb0

    when running on a ci, the libusb0 backend of pyusb can crash if access
    to the usbfs is denied: https://github.com/pyusb/pyusb/issues/151

    this fixture is automatically loaded when the CI env var is set.
    """
    with mock.patch('usb.backend.libusb0.get_backend'):
        yield





from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeRevisionFeature(SeaBreezeFeature):
    identifier = 'revision'

    def hardware_revision(self):
        raise NotImplementedError("implement in derived class")

    def revision_firmware_get(self):
        raise NotImplementedError("implement in derived class")



"""Implementation of the Protocol layer

ocean optics spectrometers use two different communication protocols:
- one is called 'OBP protocol'  # OceanBinaryProtocol
- the other one 'OOI protocol'  # ??? OceanOpticsInterface ??? maybe

"""
import collections
import functools
import hashlib
import struct
import time
import warnings

from seabreeze.pyseabreeze.exceptions import SeaBreezeError


class ProtocolInterface(object):

    def __init__(self, transport):
        if not (hasattr(transport, 'write')
                and isinstance(transport.write, collections.Callable)
                and hasattr(transport, 'read')
                and isinstance(transport.read, collections.Callable)):
            raise TypeError("transport does not implement read and write methods")
        self.transport = transport

    def send(self, msg_type, payload, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")

    def receive(self, size=None, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")

    def query(self, msg_type, payload, timeout_ms=None, **kwargs):
        raise NotImplementedError("implement in derived classes")


class OOIProtocol(ProtocolInterface):

    msgs = {code: functools.partial(struct.Struct(msg).pack, code) for code, msg in {
        0x01: '<B',    # OP_INITIALIZE
        0x02: '<BI',   # OP_ITIME
        0x05: '<BB',   # OP_GETINFO
        0x09: '<B',    # OP_REQUESTSPEC
        0x0A: '<BH',   # OP_SETTRIGMODE
        0x71: '<BBB',  # OP_TECENABLE_QE
        0x72: '<B',    # OP_READTEC_QE
        0x73: '<Bh',   # OP_TECSETTEMP_QE
        0xFE: '<B',    # OP_USBMODE
    }.items()}  # add more here if you implement new features

    def __init__(self, transport):
        super(OOIProtocol, self).__init__(transport)
        # initialize the spectrometer
        self.send(0x01)
        time.sleep(0.1)  # wait shortly after init command

    def send(self, msg_type, payload=(), timeout_ms=None, **kwargs):
        """send a ooi message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OOIProtocol.msgs`
        payload : optional
            dependent on `msg_type`. a singe value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)
        return self.transport.write(data, timeout_ms=timeout_ms)

    def receive(self, size=None, timeout_ms=None, mode=None, **kwargs):
        """receive data from the spectrometer

        Parameters
        ----------
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode : str, optional
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self.transport.read(size=size, timeout_ms=timeout_ms, mode=mode, **kwargs)

    def query(self, msg_type, payload, size=None, timeout_ms=None, mode=None, **kwargs):
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OOIProtocol.msgs`
        payload :
            the payload to be sent. Can be a singe value or a tuple, dependent
            `msg_type`.
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        mode : str, optional
            transport layers can support different modes
            (i.e. {'low_speed', 'high_speed', 'high_speed_alt'} in the usb case)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        self.send(msg_type, payload, timeout_ms=timeout_ms)
        return self.receive(size=size, timeout_ms=timeout_ms, **kwargs)


class OBPProtocol(ProtocolInterface):

    msgs = {code: struct.Struct(msg).pack for code, msg in {
        0x00000100: "",    # GET_SERIAL
        0x00100928: "",    # GET_BUF_SPEC32_META
        0x00101100: "",    # GET_RAW_SPECTRUM_NOW
        0x00110010: "<L",  # SET_ITIME_USEC
        0x00110110: "<B",  # SET_TRIG_MODE
        0x00180100: "",    # GET_WL_COEFF_COUNT
        0x00180101: "<B",  # GET_WL_COEFF
        0x00181100: "",    # GET_NL_COEFF_COUNT
        0x00181101: "<B",  # GET_NL_COEFF
        0x00420004: "",    # GET_TE_TEMPERATURE
        0x00420010: "<B",  # SET_TE_ENABLE
        0x00420011: "<f",  # SET_TE_SETPOINT
    }.items()}  # add more here if you implement new features

    class OBP(object):
        """All relevant constants are stored here"""

        HEADER_START_BYTES = 0xC0C1
        HEADER_PROTOCOL_VERSION = 0x1100  # XXX: seems to be the newest protocol version

        FLAG_RESPONSE_TO_REQUEST = 0x0001
        FLAG_ACK = 0x0002
        FLAG_REQUEST_ACK = 0x0004
        FLAG_NACK = 0x0008
        FLAG_HW_EXCEPTION = 0x0010
        FLAG_PROTOCOL_DEPRECATED = 0x0020

        ERROR_CODES = {
            0: 'Success (no detectable errors)',
            1: 'Invalid/unsupported protocol',
            2: 'Unknown message type',
            3: 'Bad checksum',
            4: 'Message too large',
            5: 'Payload length does not match message type',
            6: 'Payload data invalid',
            7: 'Device not ready for given message type',
            8: 'Unknown checksum type',
            9: 'Device reset unexpectedly',
            10: 'Too many buses (Commands have come from too many bus interfaces)',
            11: 'Out of memory. Failed to allocate enough space to complete request.',
            12: 'Command is valid, but desired information does not exist.',
            13: 'Int Device Error. May be unrecoverable.',
            100: 'Could not decrypt properly',
            101: 'Firmware layout invalid',
            102: 'Data packet was wrong size',
            103: 'hardware revision not compatible with firmware',
            104: 'Existing flash map not compatible with firmware',
            255: 'Operation/Response Deferred. Will take some time. Do not ACK or NACK yet.',
        }

        NO_ERROR = 0x0000
        RESERVED = b""
        CHECKSUM_TYPE_NONE = 0x00
        CHECKSUM_TYPE_MD5 = 0x01
        NO_CHECKSUM = b""
        FOOTER = 0xC2C3C4C5  # the datasheet specifies it in this order...

        HEADER_FMT = (
            "<H"   # start_bytes
            "H"    # protocol_version
            "H"    # flags
            "H"    # error number
            "L"    # message type
            "L"    # regarding
            "6s"   # reserved
            "B"    # checksum type
            "B"    # immediate length
            "16s"  # immediate data
            "L"    # bytes remaining
        )

        FOOTER_FMT = (
            "16s"  # checksum
            "L"    # footer
        )

    # noinspection DuplicatedCode
    def send(self, msg_type, payload=(), timeout_ms=None, request_ack=True, **kwargs):
        """send a obp message to the spectrometer

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload : optional
            dependent on `msg_type`. Can be a single value or a tuple of multiple values
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        request_ack : bool, default `True`
            request an ack for the sent command from the spectrometer.
        **kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        bytes_written : int
            the number of bytes sent
        """
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        payload = payload if isinstance(payload, (tuple, list)) else (payload,)
        data = self.msgs[msg_type](*payload)

        # Constructing message and querying usb.
        message = self._construct_outgoing_message(msg_type, data, request_ack=request_ack)  # TODO
        bytes_written = self.transport.write(message)  # ? assert bytes_written == len(message)

        if not request_ack:
            return bytes_written

        response = self.transport.read(timeout_ms=timeout_ms)
        remaining_bytes, checksum_type = self._check_incoming_message_header(response[:44])
        # ? assert remaining_bytes == 20
        checksum = self._check_incoming_message_footer(response[-20:])

        if checksum_type == self.OBP.CHECKSUM_TYPE_MD5 and checksum != hashlib.md5(response[:-20]).digest():
            warnings.warn("WARNING OBP: The checksums differ, but we ignore this for now.")
        return bytes_written

    def receive(self, size=None, timeout_ms=None, **kwargs):
        """receive data from the spectrometer

        Parameters
        ----------
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        response = self.transport.read(size=64, timeout_ms=timeout_ms)
        remaining_bytes, checksum_type = self._check_incoming_message_header(response[:44])
        length_payload_footer = remaining_bytes

        # we already received some data
        remaining_bytes -= len(response[44:])
        if remaining_bytes > 0:
            response += self.transport.read(size=remaining_bytes, timeout_ms=timeout_ms)

        if length_payload_footer != len(response[44:]):
            raise SeaBreezeError("remaining packet length mismatch: {:d} != {:d}".format(
                remaining_bytes, len(response[44:])
            ))

        checksum = self._check_incoming_message_footer(response[-20:])
        if checksum_type == self.OBP.CHECKSUM_TYPE_MD5 and checksum != hashlib.md5(response[:-20]).digest():
            warnings.warn("WARNING OBP: The checksums differ, but we ignore this for now.")

        return self._extract_message_data(response)

    def query(self, msg_type, payload=(), size=None, timeout_ms=None, **kwargs):
        """convenience method combining send and receive

        Parameters
        ----------
        msg_type : int
            a message type as defined in `OBPProtocol.msgs`
        payload :
            the payload to be sent. Can be a single value or a tuple, dependent
            `msg_type`.
        size : int, optional
            number of bytes to receive. if `None` (default) uses the
            default size as specified in the transport layer.
        timeout_ms : int, optional
            the timeout after which the transport layer should error.
            `None` means no timeout (default)
        kwargs :
            ignored and only present to provide compatible caller interfaces

        Returns
        -------
        data : str
            data returned from the spectrometer
        """
        self.send(msg_type, payload, request_ack=False)
        return self.receive(timeout_ms=timeout_ms)

    def _construct_outgoing_message(self, msg_type, payload_string, request_ack=False, regarding=None):
        """construct an outgoing OBP message

        Parameters
        ----------
        msg_type : `int`
            the obp message type, a 4 byte integer
        payload_string : `bytes`
            a compiled payload_string
        request_ack : `bool`
            request an ack for the sent command from the spectrometer.
        regarding : `int`
            see ocean optics obp protocol documentation

        Returns
        -------
        message : `str`
            compiled message
        """
        if request_ack is True:
            flags = self.OBP.FLAG_REQUEST_ACK
        else:
            flags = 0

        if regarding is None:
            regarding = 0

        if len(payload_string) <= 16:
            payload_string_fmt = "0s"
            immediate_length = len(payload_string)
            immediate_data = payload_string
            payload_string = b""
            bytes_remaining = 20  # Checksum + footer
        else:
            payload_string_fmt = "%ds" % len(payload_string)
            immediate_length = 0
            immediate_data = b""
            bytes_remaining = 20 + len(payload_string)

        FMT = self.OBP.HEADER_FMT + payload_string_fmt + self.OBP.FOOTER_FMT

        msg = struct.pack(FMT,
                          self.OBP.HEADER_START_BYTES,
                          self.OBP.HEADER_PROTOCOL_VERSION,
                          flags,
                          self.OBP.NO_ERROR,
                          msg_type,
                          regarding,
                          self.OBP.RESERVED,
                          self.OBP.CHECKSUM_TYPE_NONE,
                          immediate_length,
                          immediate_data,
                          bytes_remaining,
                          payload_string,
                          self.OBP.NO_CHECKSUM,
                          self.OBP.FOOTER)
        return msg

    def _check_incoming_message_header(self, header):
        """check the incoming message header

        Parameters
        ----------
        header : `str`
            a obp header of length 44

        Returns
        -------
        bytes_and_checksum_type : tuple[`int`, `int`]
            bytes_remaining after the header (returns 20 for a 64 byte message)
            checksum_type only supports self.OBP.CHECKSUM_TYPE_MD5 for now
        """
        if len(header) != 44:
            raise SeaBreezeError("header has wrong length! len(header): %d" % len(header))

        data = struct.unpack(self.OBP.HEADER_FMT, header)

        if data[0] != self.OBP.HEADER_START_BYTES:
            raise SeaBreezeError('Header start_bytes wrong: "%d"' % data[0])
        if data[1] != self.OBP.HEADER_PROTOCOL_VERSION:
            raise SeaBreezeError('Header protocol version wrong: %d' % data[1])

        flags = data[2]
        if flags == 0:
            pass
        if flags & self.OBP.FLAG_RESPONSE_TO_REQUEST:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_ACK:
            pass  # TODO: propagate?
        if flags & self.OBP.FLAG_REQUEST_ACK:
            pass  # TODO: only the host should be able to set this?
        if (flags & self.OBP.FLAG_NACK) or (flags & self.OBP.FLAG_HW_EXCEPTION):
            error = data[3]
            if error != 0:  # != SUCCESS
                raise SeaBreezeError(self.OBP.ERROR_CODES[error])
            else:
                pass  # TODO: should we do something here?
        if flags & self.OBP.FLAG_PROTOCOL_DEPRECATED:
            raise SeaBreezeError("Protocol deprecated?!?")

        # msg_type = data[4]
        # regarding = data[5]

        checksum_type = data[7]  # TODO: implement checksums.
        if checksum_type not in [self.OBP.CHECKSUM_TYPE_NONE, self.OBP.CHECKSUM_TYPE_MD5]:
            raise SeaBreezeError('the checksum type is unknown: "%d"' % checksum_type)

        # immediate_length = data[8]
        # immediate_data = data[9]
        bytes_remaining = data[10]

        return bytes_remaining, checksum_type

    def _check_incoming_message_footer(self, footer):
        """check the incoming message header

        Parameters
        ----------
        footer : `str`
            a obp footer of length 20

        Returns
        -------
        checksum: `str`
            the 16 byte checksum of the message
        """
        assert len(footer) == 20, "footer has wrong length! len(footer): %d" % len(footer)

        data = struct.unpack("<" + self.OBP.FOOTER_FMT, footer)

        checksum = data[0]
        assert data[1] == self.OBP.FOOTER, "the device returned a wrong footer: %d" % data[1]

        return checksum

    def _extract_message_data(self, msg):
        """extract the payload data from a obp message

        Parameters
        ----------
        msg : `str`
            a obp message

        Returns
        -------
        data : `str`
            the payload contained in the message
        """
        payload_length = len(msg) - 44 - 20  # - HeaderLength - FooterLength
        if not (payload_length >= 0):
            raise SeaBreezeError("Received message < 64 bytes: %d" % payload_length)
        payload_fmt = "%ds" % payload_length
        FMT = self.OBP.HEADER_FMT + payload_fmt + self.OBP.FOOTER_FMT

        data = struct.unpack(FMT, msg)

        msg_type = data[4]

        immediate_length = data[8]
        immediate_data = data[9]
        payload = data[11]

        if (immediate_length > 0) and len(payload) > 0:
            raise SeaBreezeError("Got immediate AND payload data? cmd: '%d'" % msg_type)
        elif immediate_length > 0:
            return immediate_data[:immediate_length]
        elif payload_length > 0:
            return payload
        else:
            return b""



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeContinuousStrobeFeature(SeaBreezeFeature):
    identifier = 'continuous_strobe'

    def set_enable(self, strobe_enable):
        raise NotImplementedError("implement in derived class")

    def set_period_micros(self, period_micros):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeIPv4Feature(SeaBreezeFeature):
    identifier = 'ipv4'

    def get_dhcp_enable_state(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_dhcp_enable_state(self, interface_index, is_enabled):
        raise NotImplementedError("implement in derived class")

    def get_number_of_ipv4_addresses(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def get_ipv4_address(self, interface_index, address_index):
        raise NotImplementedError("implement in derived class")

    def get_default_gateway(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_default_gateway(self, interface_index, default_gateway_address):
        raise NotImplementedError("implement in derived class")

    def add_ipv4_address(self, interface_index, ipv4_address):
        raise NotImplementedError("implement in derived class")

    def delete_ipv4_address(self, interface_index, address_index):
        raise NotImplementedError("implement in derived class")



"""This is a wrapper for the pyusb-implementation of the seabreeze-library

"""
from seabreeze.pyseabreeze.api import SeaBreezeAPI
from seabreeze.pyseabreeze.exceptions import (SeaBreezeError,
                                              SeaBreezeNumFeaturesError)
from seabreeze.pyseabreeze.devices import SeaBreezeDevice
# noinspection PyUnresolvedReferences,PyProtectedMember
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
# noinspection PyUnresolvedReferences
from seabreeze.pyseabreeze.features import (SeaBreezeRawUSBBusAccessFeature,
                                            SeaBreezeSpectrometerFeature,
                                            SeaBreezePixelBinningFeature,
                                            SeaBreezeThermoElectricFeature,
                                            SeaBreezeIrradCalFeature,
                                            SeaBreezeEthernetConfigurationFeature,
                                            SeaBreezeMulticastFeature,
                                            SeaBreezeIPv4Feature,
                                            SeaBreezeDHCPServerFeature,
                                            SeaBreezeNetworkConfigurationFeature,
                                            SeaBreezeWifiConfigurationFeature,
                                            SeaBreezeGPIOFeature,
                                            SeaBreezeEEPROMFeature,
                                            SeaBreezeLightSourceFeature,
                                            SeaBreezeStrobeLampFeature,
                                            SeaBreezeContinuousStrobeFeature,
                                            SeaBreezeShutterFeature,
                                            SeaBreezeNonlinearityCoefficientsFeature,
                                            SeaBreezeTemperatureFeature,
                                            SeaBreezeIntrospectionFeature,
                                            SeaBreezeSpectrumProcessingFeature,
                                            SeaBreezeRevisionFeature,
                                            SeaBreezeOpticalBenchFeature,
                                            SeaBreezeStrayLightCoefficientsFeature,
                                            SeaBreezeDataBufferFeature,
                                            SeaBreezeFastBufferFeature,
                                            SeaBreezeAcquisitionDelayFeature,
                                            SeaBreezeI2CMasterFeature)
_backend_ = 'pyseabreeze'



"""Implementation of the Seabreeze Transport layer.

Some spectrometers can support different transports (usb, network, rs232, etc.)

"""
import warnings
from functools import partial

import usb.core
import usb.util

try:
    from functools import partialmethod
except ImportError:
    # https://gist.github.com/carymrobbins/8940382
    # noinspection PyPep8Naming
    class partialmethod(partial):
        def __get__(self, instance, owner):
            if instance is None:
                return self
            args, kwargs = self.args or (), self.keywords or {}
            return partial(self.func, instance, *args, **kwargs)


class TransportInterface(object):

    _required_init_kwargs = ()

    def open_device(self, device):
        """open a seabreeze device

        Parameters
        ----------
        device : seabreeze.pyseabreeze.devices.SeaBreezeDevice
        """
        raise NotImplementedError("implement in derived transport class")

    @property
    def is_open(self):
        """return if device is opened

        Returns
        -------
        bool
        """
        raise NotImplementedError("implement in derived transport class")

    def close_device(self):
        """close the seabreeze device"""
        raise NotImplementedError("implement in derived transport class")

    def write(self, data, timeout_ms=None, **kwargs):
        """write data to the device"""
        raise NotImplementedError("implement in derived transport class")

    def read(self, size=None, timeout_ms=None, **kwargs):
        """read data from the

        Returns
        -------
        str
        """
        raise NotImplementedError("implement in derived transport class")

    @property
    def default_timeout_ms(self):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def list_devices(cls):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def register_model(cls, model_name, **kwargs):
        raise NotImplementedError("implement in derived transport class")

    @classmethod
    def supported_model(cls, device):
        """return if the device supports the transport or vice versa

        Returns
        -------
        model_name : ``
        """
        return None

    @classmethod
    def specialize(cls, model_name, **kwargs):
        raise NotImplementedError("implement in derived transport class")


class USBTransport(TransportInterface):
    """implementation of the usb transport interface for spectrometers"""

    _required_init_kwargs = ('usb_product_id', 'usb_endpoint_map', 'usb_protocol')
    vendor_id = 0x2457
    product_ids = {}

    def __init__(self, usb_product_id, usb_endpoint_map, usb_protocol):
        super(USBTransport, self).__init__()
        self._product_id = usb_product_id
        self._endpoint_map = usb_endpoint_map
        self._protocol_cls = usb_protocol
        # internal settings
        self._default_read_size = {
            'low_speed': 64,
            'high_speed': 512,
            'high_speed_alt': 512
        }
        self._read_endpoints = {
            'low_speed': 'lowspeed_in',
            'high_speed': 'highspeed_in',
            'high_speed_alt': 'highspeed_in2'
        }
        self._default_read_endpoint = 'low_speed'
        # internal state
        self._device = None
        self._opened = None
        self._protocol = None

    def open_device(self, device):
        if not isinstance(device, usb.core.Device):
            raise TypeError("device needs to be usb.core.Device")
        # device.reset()
        self._device = device
        try:
            if self._device.is_kernel_driver_active(0):
                self._device.detach_kernel_driver(0)
        except NotImplementedError:
            pass  # unavailable on some systems/backends
        try:
            self._device.set_configuration()
        except usb.core.USBError as err:
            if err.errno == 16:
                # TODO: warn as in cseabreeze
                self._opened = True
            raise
        else:
            self._opened = True
        # This will initialize the communication protocol
        if self._opened:
            self._protocol = self._protocol_cls(self)

    @property
    def is_open(self):
        return self._opened or False

    def close_device(self):
        usb.util.dispose_resources(self._device)
        self._device.reset()
        self._device = None
        self._opened = False
        self._protocol = None

    def write(self, data, timeout_ms=None, **kwargs):
        if self._device is None:
            raise RuntimeError("device not opened")
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self._device.write(self._endpoint_map.ep_out, data, timeout=timeout_ms)

    def read(self, size=None, timeout_ms=None, mode=None, **kwargs):
        if self._device is None:
            raise RuntimeError("device not opened")
        mode = mode if mode is not None else self._default_read_endpoint
        endpoint = getattr(self._endpoint_map, self._read_endpoints[mode])
        if size is None:
            size = self._default_read_size[mode]
        if kwargs:
            warnings.warn("kwargs provided but ignored: {}".format(kwargs))
        return self._device.read(endpoint, size, timeout=timeout_ms)

    @property
    def default_timeout_ms(self):
        return self._device.default_timeout if self._device else None

    @classmethod
    def list_devices(cls):
        """list pyusb devices for all available spectrometers

        Note: this includes spectrometers that are currently opened in other
        processes on the machine.

        Returns
        -------
        devices : list of usb.core.Device
            unique pyusb devices for each available spectrometer
        """
        # get all matching devices
        return usb.core.find(find_all=True,
                             custom_match=lambda dev: (dev.idVendor == cls.vendor_id and
                                                       dev.idProduct in cls.product_ids))

    @classmethod
    def register_model(cls, model_name, **kwargs):
        product_id = kwargs.get('usb_product_id')
        if product_id in cls.product_ids:
            raise ValueError("product_id 0x{:04x} already in registry".format(product_id))
        cls.product_ids[product_id] = model_name

    @classmethod
    def supported_model(cls, device):
        if not isinstance(device, usb.core.Device):
            return None
        # noinspection PyUnresolvedReferences
        return cls.product_ids[device.idProduct]

    @classmethod
    def specialize(cls, model_name, **kwargs):
        assert set(kwargs) == set(cls._required_init_kwargs)
        # usb transport register automatically on registration
        cls.register_model(model_name, **kwargs)
        specialized_class = type("USBTransport{}".format(model_name), (cls, ), {
                                    '__init__': partialmethod(cls.__init__, **kwargs)
                                 })
        return specialized_class



"""This is the pyusb implementation of the seabreeze library

Author: Andreas Poehlmann

most of this code is based on implementations from
    http://github.com/ap--/python-oceanoptics

"""
import logging

import usb.core
import usb.util

from seabreeze.pyseabreeze.devices import SeaBreezeDevice, _model_class_registry
from seabreeze.pyseabreeze.transport import USBTransport


__all__ = ['SeaBreezeAPI']


class SeaBreezeAPI(object):
    """SeaBreeze API interface"""

    _log = logging.getLogger(__name__)

    def __init__(self, initialize=True):
        if initialize:
            self.initialize()

    @staticmethod
    def initialize():
        """initialize the api backend

        normally this function does not have to be called directly by the user.
        it resets all usb devices on load
        """
        # create the handles and associate protocols
        for device in USBTransport.list_devices():
            try:
                device.reset()
                usb.util.dispose_resources(device)
            except Exception as err:
                SeaBreezeAPI._log.debug("initialize failed: {}('{}')".format(
                    err.__class__.__name__, err.message
                ))

    @staticmethod
    def shutdown():
        """shutdown the api backend

        normally this function does not have to be called directly by the user
        """
        # dispose usb resources
        for device in USBTransport.list_devices():
            try:
                usb.util.dispose_resources(device)
            except Exception as err:
                SeaBreezeAPI._log.debug("shutdown failed: {}('{}')".format(
                    err.__class__.__name__, err.message
                ))

    @staticmethod
    def add_rs232_device_location(device_type, bus_path, baudrate):
        """add RS232 device location"""
        # RS232Transport.register_device(device_type, bus_path, baudrate)
        raise NotImplementedError("rs232 communication not implemented for pyseabreeze")

    @staticmethod
    def add_ipv4_device_location(device_type, ip_address, port):
        """add ipv4 device location"""
        # IPV4Transport.register_device(device_type, ip_address, port)
        raise NotImplementedError("ipv4 communication not implemented for pyseabreeze")

    @staticmethod
    def list_devices():
        """returns available SeaBreezeDevices

        list all connected Ocean Optics devices supported
        by libseabreeze.

        Returns
        -------
        devices: list of SeaBreezeDevice
            connected Spectrometer instances
        """
        # get all matching devices
        devices = []
        for pyusb_dev in USBTransport.list_devices():
            # get the correct communication interface
            dev = SeaBreezeDevice(pyusb_dev)
            if dev.is_open:
                was_open_before = True
            else:
                was_open_before = False
                try:
                    dev.open()
                except usb.core.USBError as usb_error:
                    if usb_error.errno == 16:
                        # device used by another thread?
                        continue
                    else:
                        raise
            if not was_open_before:
                dev.close()
            devices.append(dev)
        return devices

    @staticmethod
    def supported_models():
        """returns SeaBreezeDevices supported by the backend

        models supported by the backend

        Returns
        -------
        devices: list of str
            list of model names that are supported by this backend
        """
        return [x for x in sorted(_model_class_registry.keys())]



"""
SeaBreeze
=========

Python module for accessing Ocean Optics Spectrometers.
Returns spectra as NumPy Arrays and unifies the spectrometer
control for all supported spectrometer models.

How to use the module
---------------------
To measure a spectrum do::

  >>> import seabreeze
  >>> seabreeze.use("cseabreeze")
  >>> from seabreeze.spectrometers import Spectrometer
  >>> spec = Spectrometer.from_serial_number(serial=None)  # serial=None picks the first available
  >>> wavelengths = spec.wavelengths()
  >>> intensities = spec.intensities()

Available subpackages
---------------------
spectrometers
    Core interface to the spectrometers
backend
    internal backend loading functions
cseabreeze
    libseabreeze Cython wrapped backend
pyseabreeze
    pyusb implementation of the cseabreeze backend

Utilities
---------
utils
    -
system
    operating system related functionality

Operating System dependent setup
--------------------------------
Dependent on your operating system there's steps you have to take to make
the spectrometer visible to python-seabreeze.

Linux
    requires to install the udev rules
Osx
    should work ootb
Windows
    requires installing the drivers

"""
# noinspection PyUnresolvedReferences
from seabreeze.backends import use
from seabreeze._version import __version__

__author__ = "Andreas Poehlmann"
__email__ = "andreas@poehlmann.io"



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeSpectrumProcessingFeature(SeaBreezeFeature):
    identifier = 'spectrum_processing'

    def get_boxcar_width(self):
        raise NotImplementedError("implement in derived class")

    def set_boxcar_width(self, boxcar_width):
        raise NotImplementedError("implement in derived class")

    boxcar_width = property(get_boxcar_width, set_boxcar_width)

    def get_scans_to_average(self):
        raise NotImplementedError("implement in derived class")

    def set_scans_to_average(self, scans_to_average):
        raise NotImplementedError("implement in derived class")

    scans_to_average = property(get_scans_to_average, set_scans_to_average)



import logging
import warnings

__all__ = ['use', 'get_backend']

_log = logging.getLogger(__name__)


def _use_backend(name):
    # internal: import the libseabreeze cython wrapper -> cseabreeze
    try:
        if name == 'cseabreeze':
            import seabreeze.cseabreeze as sbb
        elif name == 'pyseabreeze':
            import seabreeze.pyseabreeze as sbb
        else:
            raise ValueError("unknown backend '{}'".format(name))
    except ImportError as err:
        _log.warn("seabreeze can't load '{}' backend - error: '{}'".format(
            name, repr(err)
        ), exc_info=True)
        return None
    else:
        return sbb


# internal config dict used exclusively in this submodule
_SeaBreezeConfig = {
    'requested_backend': 'cseabreeze',  # default is cseabreeze
    'available_backends': ('cseabreeze', 'pyseabreeze'),
    'allow_fallback': False
}


def use(backend, force=True):
    """
    select the backend used for communicating with the spectrometer

    Parameters
    ----------
    backend : str

    force : bool, optional, default: True
        raises an ImportError when ``seabreeze.get_backend()`` is called
        and the requested backend can not be imported. force=True should
        be used in user code to ensure that the correct backend is being
        loaded.
    """
    if backend not in _SeaBreezeConfig['available_backends']:
        raise ValueError('backend not in: {}'.format(', '.join(_SeaBreezeConfig['available_backends'])))
    _SeaBreezeConfig['requested_backend'] = backend
    _SeaBreezeConfig['allow_fallback'] = not force


def get_backend():
    """
    return the requested backend or a fallback. configuration is done
    via ``seabreeze.use()``

    Returns
    -------
    backend:
        a backend interface for communicating with the spectrometers
    """
    requested = _SeaBreezeConfig['requested_backend']
    fallback = _SeaBreezeConfig['allow_fallback']
    backends = _SeaBreezeConfig['available_backends']

    backend = _use_backend(requested)  # trying to import requested backend
    if backend is None and fallback:

        warnings.warn("seabreeze failed to load user requested '{}' backend but will try fallback backends".format(
            requested
        ), category=ImportWarning)

        fallback_backends = [b for b in backends if b != requested]
        for fallback_backend in fallback_backends:
            backend = _use_backend(fallback_backend)
            if backend is not None:
                break
        else:
            raise ImportError("Could not import backend. Tried: {}".format(
                ', '.join(backends)
            ))

    if backend is None:
        raise ImportError("Could not import backend. Requested: {}".format(requested))

    return backend



"""
seabreeze.spectrometers core module
===================================

Author: Andreas Poehlmann
Email: andreas@poehlmann.io

"""
import numpy

import seabreeze.backends
from seabreeze.compat import DeprecatedSpectrometerMixin as _DeprecatedSpectrometerMixin

# get the backend and add some functions/classes to this module
_lib = seabreeze.backends.get_backend()

SeaBreezeError = _lib.SeaBreezeError
SeaBreezeDevice = _lib.SeaBreezeDevice
SeaBreezeFeature = _lib.SeaBreezeFeature


def list_devices():
    """returns available SeaBreezeDevices

    list all connected Ocean Optics devices supported

    Returns
    -------
    devices: `list[SeaBreezeDevice]`
        connected Spectrometer instances
    """
    if not hasattr(list_devices, '_api'):
        list_devices._api = _lib.SeaBreezeAPI()
    # noinspection PyProtectedMember
    return list_devices._api.list_devices()


class Spectrometer(_DeprecatedSpectrometerMixin):
    """Spectrometer class for all supported spectrometers"""

    _backend = _lib  # store reference to backend to allow backend switching in tests

    def __init__(self, device):
        """create a Spectrometer instance for the provided device

        The Spectrometer class provides a thin abstraction layer for the
        basic spectrometer feature of the provided SeaBreezeDevice.

        Parameters
        ----------
        device : `seabreeze.spectrometers.SeaBreezeDevice`
            a SeaBreezeDevice as returned from `list_devices`
        """
        if not isinstance(device, self._backend.SeaBreezeDevice):
            raise TypeError("device has to be a `SeaBreezeDevice`")
        self._dev = device
        self.open()  # always open the device here to allow caching values

        # check for nonlinearity correction support
        nc_feature = self._dev.f.nonlinearity_coefficients
        self._nc = None
        if nc_feature is not None:
            try:
                # NOTE: the spark spectrometer raises a transport error when trying
                # to receive the nc coefficients. In this case continue with disabled
                # nonlinearity correction support
                self._nc = nc_feature.get_nonlinearity_coefficients()
            except self._backend.SeaBreezeError:
                pass
        # check for dark pixel correction support
        self._dp = self._dev.f.spectrometer.get_electric_dark_pixel_indices()
        # cache wavelengths on open
        self._wavelengths = self._dev.f.spectrometer.get_wavelengths()

    @classmethod
    def from_first_available(cls):
        """open first available spectrometer

        Returns
        -------
        spectrometer : `Spectrometer`
            the first available supported spectrometer
        """
        for dev in list_devices():
            if not dev.is_open:
                return cls(dev)
        else:
            raise cls._backend.SeaBreezeError("No unopened device found.")

    @classmethod
    def from_serial_number(cls, serial=None):
        """open the spectrometer matching the provided serial number

        Allows to open a specific spectrometer if multiple are connected.
        Spectrometer serial numbers are visible in the repr string of
        each `SeaBreezeDevice` or their `serial_number` attribute.

        Parameters
        ----------
        serial : `str`, optional
            the spectrometer's serial number. If `None` (default) it
            returns the first available unopened spectrometer.

        Returns
        -------
        spectrometer : `Spectrometer`
            the spectrometer with the requested serial number
        """
        if serial is None:  # pick first spectrometer
            return cls.from_first_available()

        for dev in list_devices():
            if dev.serial_number == str(serial):
                if dev.is_open:
                    raise cls._backend.SeaBreezeError("Device already opened.")
                else:
                    return cls(dev)
        else:
            raise cls._backend.SeaBreezeError("No device attached with serial number '%s'." % serial)

    def wavelengths(self):
        """wavelength array of the spectrometer

        wavelengths in (nm) corresponding to each pixel of the spectrometer

        Returns
        -------
        wavelengths : `numpy.ndarray`
            wavelengths in (nm)
        """
        return self._wavelengths

    def intensities(self, correct_dark_counts=False, correct_nonlinearity=False):
        """measured intensity array in (a.u.)

        Measured intensities as numpy array returned by the spectrometer.
        The measuring behavior can be adjusted by setting the trigger mode.
        Pixels at the start and end of the array might not be optically
        active so interpret their returned measurements with care. Refer
        to the spectrometer's datasheet for further information.

        Notes
        -----
        Intensities are in arbitrary units and the range depends on the
        ADC bit resolution of the hardware used in the specific spectrometer.
        Some spectrometers store a `saturation` value in their eeprom,
        which is used to rescale the raw ADC output to the full bit range.
        (This is done in `libseabreeze` and therefore also in `cseabreeze`
        --- for compatibility reasons the same is done in `pyseabreeze`)
        I.e. this means that a 16bit (max value 65535) spectrometer with a
        saturation value of ~30000 is effectively only returning ~15bit
        resolution raw readings. While most of the lower bits are dominated
        by noise anyways, it's just something to keep in mind. Refer to
        `pyseabreeze.features.spectrometer._SeaBreezeSpectrometerSaturationMixin`
        for the implementation.

        Parameters
        ----------
        correct_dark_counts : `bool`
            If requested and supported the average value of electric dark
            pixels on the ccd of the spectrometer is subtracted from the
            measurements to remove the noise floor in the measurements
            caused by non optical noise sources.
        correct_nonlinearity : `bool`
            Some spectrometers store non linearity correction coefficients
            in their eeprom. If requested and supported by the spectrometer
            the readings returned by the spectrometer will be linearized
            using the stored coefficients.

        Returns
        -------
        intensities : `numpy.ndarray`
            measured intensities in (a.u.)
        """
        if correct_dark_counts and not self._dp:
            raise self._backend.SeaBreezeError("This device does not support dark count correction.")
        if correct_nonlinearity and not self._nc:
            raise self._backend.SeaBreezeError("This device does not support nonlinearity correction.")
        # Get the intensities
        out = self._dev.f.spectrometer.get_intensities()
        # Do corrections if requested
        if correct_nonlinearity or correct_dark_counts:
            dark_offset = numpy.mean(out[self._dp]) if self._dp else 0.
            out -= dark_offset
        if correct_nonlinearity:
            out = out / numpy.polyval(self._nc, out)
        if correct_nonlinearity and (not correct_dark_counts):
            # noinspection PyUnboundLocalVariable
            out += dark_offset
        return out

    @property
    def max_intensity(self):
        """return the maximum intensity of the spectrometer

        Returns
        -------
        max_intensity : `float`
            the maximum intensity that can be returned by the spectrometer in (a.u.)
            It's possible that the spectrometer saturates already at lower values.
        """
        return self._dev.f.spectrometer.get_maximum_intensity()

    def spectrum(self, correct_dark_counts=False, correct_nonlinearity=False):
        """returns wavelengths and intensities as single array

        Convenience method to allow:

        >>> spec = Spectrometer.from_first_available()
        >>> wavelengths, intensities = spec.spectrum()

        Parameters
        ----------
        correct_dark_counts : `bool`
            see `Spectrometer.intensities`
        correct_nonlinearity : `bool`
            see `Spectrometer.intensities`

        Returns
        -------
        spectrum : `numpy.ndarray`
            combined array of wavelengths and measured intensities
        """
        return numpy.vstack((self._wavelengths,
                             self.intensities(correct_dark_counts, correct_nonlinearity)))

    def integration_time_micros(self, integration_time_micros):
        """set the integration time in microseconds

        Parameters
        ----------
        integration_time_micros : `int`
            integration time in microseconds
        """
        # NOTE: Protect against a bug in libseabreeze:
        # If integration time is out of bounds, libseabreeze returns Undefined Error
        # (Probably only for devices with a non micro second time base...)
        try:
            self._dev.f.spectrometer.set_integration_time_micros(integration_time_micros)
        except OverflowError:
            raise self._backend.SeaBreezeError("[OverFlow] Specified integration time is out of range.")
        except self._backend.SeaBreezeError as e:
            if getattr(e, 'error_code', None) == 1:
                # Only replace if 'Undefined Error'
                raise self._backend.SeaBreezeError("[Fix] Specified integration time is out of range.")
            else:
                raise e

    @property
    def integration_time_micros_limits(self):
        """return the hardcoded minimum and maximum integration time

        Returns
        -------
        integration_time_micros_min_max : `tuple[int, int]`
            min and max integration time in micro seconds
        """
        return self._dev.f.spectrometer.get_integration_time_micros_limits()

    def trigger_mode(self, mode):
        """set the trigger mode of the device

        Parameters
        ----------
        mode : `int`
            refer to your spectrometer's datasheet to determine the correct
            value for the trigger mode you want to use.
        """
        self._dev.f.spectrometer.set_trigger_mode(mode)

    @property
    def serial_number(self):
        """the spectrometer's serial number"""
        return self._dev.serial_number

    @property
    def model(self):
        """the spectrometer's model type"""
        return self._dev.model

    @property
    def pixels(self):
        """the spectrometer's number of pixels"""
        # noinspection PyProtectedMember
        return self._dev.f.spectrometer._spectrum_length

    @property
    def features(self):
        """return a dictionary of all supported features

        this returns a dictionary with all supported Features of the spectrometer
        and gives direct access to the features provided by the backend interface.

        Returns
        -------
        features : `dict` [`str`, `seabreeze.SeaBreezeFeature`]
        """
        return self._dev.features

    @property
    def f(self):
        """convenience assess to features via attributes

        this allows you to access a feature like this::

        >>> spec = Spectrometer.from_first_available()
        >>> # via .features
        >>> spec.features['eeprom'][0].eeprom_read_slot(4)
        >>> # via .f
        >>> spec.f.eeprom.eeprom_read_slot(4)

        """
        return self._dev.f

    def open(self):
        """open the connection to the SeaBreezeDevice

        Notes
        -----
        Normally you do not have to call this function manually.
        If you're trying to use the same spectrometer from multiple
        processes (honestly, why would you?) then this might come
        in handy. But I'll leave the lock/semaphore handling to you.
        """
        self._dev.open()

    def close(self):
        """close the connection to the SeaBreezeDevice

        Notes
        -----
        Normally you do not have to call this function manually.
        If you're trying to use the same spectrometer from multiple
        processes (honestly, why would you?) then this might come
        in handy. But I'll leave the lock/semaphore handling to you.
        """
        self._dev.close()

    def __repr__(self):
        return "<Spectrometer %s:%s>" % (self.model, self.serial_number)



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
# 
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeShutterFeature(SeaBreezeFeature):
    identifier = 'shutter'

    def set_shutter_open(self, state):
        raise NotImplementedError("implement in derived class")
